@@ -1,5 +1,5 @@
 /*
- * $Id: avpair.c,v 1.23 2008/01/09 07:05:11 sobomax Exp $
+ * $Id: avpair.c,v 1.26 2010/06/15 09:22:52 aland Exp $
  *
  * Copyright (C) 1995 Lars Fenneberg
  *
@@ -29,7 +29,7 @@
  *
  */
 
-VALUE_PAIR *rc_avpair_add (const rc_handle *rh, VALUE_PAIR **list, int attrid, void *pval, int len, int vendorpec)
+VALUE_PAIR *rc_avpair_add (rc_handle const *rh, VALUE_PAIR **list, int attrid, void const *pval, int len, int vendorpec)
 {
 	VALUE_PAIR     *vp;
 
@@ -54,19 +54,19 @@ VALUE_PAIR *rc_avpair_add (const rc_handle *rh, VALUE_PAIR **list, int attrid, v
  *
  */
 
-int rc_avpair_assign (VALUE_PAIR *vp, void *pval, int len)
+int rc_avpair_assign (VALUE_PAIR *vp, void const *pval, int len)
 {
 
 	switch (vp->type)
 	{
 		case PW_TYPE_STRING:
 			if (len == -1)
-				len = (uint32_t)strlen((char *)pval);
+				len = (uint32_t)strlen((char const *)pval);
 			if (len > AUTH_STRING_LEN) {
 		        	rc_log(LOG_ERR, "rc_avpair_assign: bad attribute length");
 		        	return -1;
 			}
-			memcpy(vp->strvalue, (char *)pval, len);
+			memcpy(vp->strvalue, (char const *)pval, len);
 			vp->strvalue[len] = '\0';
 			vp->lvalue = len;
 			break;
@@ -76,6 +76,23 @@ int rc_avpair_assign (VALUE_PAIR *vp, void *pval, int len)
 	        case PW_TYPE_IPADDR:
 			vp->lvalue = * (uint32_t *) pval;
 			break;
+	        case PW_TYPE_IPV6ADDR:
+			if (len != 16) {
+		        	rc_log(LOG_ERR, "rc_avpair_assign: bad IPv6 length");
+		        	return -1;
+			}
+			memcpy(vp->strvalue, (char const *)pval, len);
+			vp->lvalue = len;
+			break;
+
+	        case PW_TYPE_IPV6PREFIX:
+			if (len < 2 || len > 18) {
+		        	rc_log(LOG_ERR, "rc_avpair_assign: bad IPv6 prefix length");
+		        	return -1;
+			}
+			memcpy(vp->strvalue, (char const *)pval, len);
+			vp->lvalue = len;
+			break;
 
 		default:
 			rc_log(LOG_ERR, "rc_avpair_assign: unknown attribute %d", vp->type);
@@ -93,7 +110,7 @@ int rc_avpair_assign (VALUE_PAIR *vp, void *pval, int len)
  *
  */
 
-VALUE_PAIR *rc_avpair_new (const rc_handle *rh, int attrid, void *pval, int len, int vendorpec)
+VALUE_PAIR *rc_avpair_new (rc_handle const *rh, int attrid, void const *pval, int len, int vendorpec)
 {
 	VALUE_PAIR     *vp = NULL;
 	DICT_ATTR      *pda;
@@ -165,11 +182,11 @@ VALUE_PAIR *rc_avpair_new (const rc_handle *rh, int attrid, void *pval, int len,
  */
 
 VALUE_PAIR *
-rc_avpair_gen(const rc_handle *rh, VALUE_PAIR *pair, unsigned char *ptr,
+rc_avpair_gen(rc_handle const *rh, VALUE_PAIR *pair, unsigned char const *ptr,
     int length, int vendorpec)
 {
 	int attribute, attrlen, x_len;
-	unsigned char *x_ptr;
+	unsigned char const *x_ptr;
 	uint32_t lvalue;
 	DICT_ATTR *attr;
 	VALUE_PAIR *rpair;
@@ -228,7 +245,7 @@ rc_avpair_gen(const rc_handle *rh, VALUE_PAIR *pair, unsigned char *ptr,
 		buffer[0] = '\0';	/* Initial length. */
 		x_ptr = ptr;
 		for (x_len = attrlen; x_len > 0; x_len--, x_ptr++) {
-			sprintf(hex, "%2.2X", x_ptr[0]);
+			snprintf(hex, sizeof(hex), "%2.2X", x_ptr[0]);
 			strcat(buffer, hex);
 		}
 		if (vendorpec == 0) {
@@ -280,6 +297,30 @@ rc_avpair_gen(const rc_handle *rh, VALUE_PAIR *pair, unsigned char *ptr,
 		memcpy((char *)&lvalue, (char *)ptr, 4);
 		pair->lvalue = ntohl(lvalue);
 		break;
+	case PW_TYPE_IPV6ADDR:
+		if (attrlen != 16) {
+			rc_log(LOG_ERR, "rc_avpair_gen: received IPV6ADDR"
+			    " attribute with invalid length");
+			goto shithappens;
+		}
+		memcpy(pair->strvalue, (char *)ptr, 16);
+		pair->lvalue = attrlen;
+		break;
+	case PW_TYPE_IPV6PREFIX:
+		if (attrlen > 18 || attrlen < 2) {
+			rc_log(LOG_ERR, "rc_avpair_gen: received IPV6PREFIX"
+			    " attribute with invalid length: %d", attrlen);
+			goto shithappens;
+		}
+		memcpy(pair->strvalue, (char *)ptr, attrlen);
+		pair->lvalue = attrlen;
+		break;
+	case PW_TYPE_DATE:
+		if (attrlen != 4) {
+			rc_log(LOG_ERR, "rc_avpair_gen: received DATE "
+			    "attribute with invalid length");
+			goto shithappens;
+		}
 
 	default:
 		rc_log(LOG_WARNING, "rc_avpair_gen: %s has unknown type",
@@ -403,9 +444,9 @@ void rc_avpair_free (VALUE_PAIR *pair)
  */
 
 static void
-rc_fieldcpy(char *string, char **uptr, const char *stopat, size_t len)
+rc_fieldcpy(char *string, char const **uptr, char const *stopat, size_t len)
 {
-	char *ptr, *estring;
+	char const *ptr, *estring;
 
 	ptr = *uptr;
 	estring = string + len - 1;
@@ -454,11 +495,11 @@ rc_fieldcpy(char *string, char **uptr, const char *stopat, size_t len)
 #define PARSE_MODE_VALUE	2
 #define PARSE_MODE_INVALID	3
 
-int rc_avpair_parse (const rc_handle *rh, char *buffer, VALUE_PAIR **first_pair)
+int rc_avpair_parse (rc_handle const *rh, char const *buffer, VALUE_PAIR **first_pair)
 {
 	int             mode;
 	char            attrstr[AUTH_ID_LEN];
-	char            valstr[AUTH_STRING_LEN + 1];
+	char            valstr[AUTH_STRING_LEN + 1], *p;
 	DICT_ATTR      *attr = NULL;
 	DICT_VALUE     *dval;
 	VALUE_PAIR     *pair;
@@ -562,6 +603,35 @@ int rc_avpair_parse (const rc_handle *rh, char *buffer, VALUE_PAIR **first_pair)
                                 pair->lvalue = rc_get_ipaddr(valstr);
 				break;
 
+			    case PW_TYPE_IPV6ADDR:
+			    	if (inet_pton(AF_INET6, valstr, pair->strvalue) == 0) {
+			    		rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv6 address %s", valstr);
+			    		free(pair);
+			    		return -1;
+			    	}
+				pair->lvalue = 16;
+				break;
+
+			    case PW_TYPE_IPV6PREFIX:
+			    	p = strchr(valstr, '/');
+			    	if (p == NULL) {
+			    		rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv6 prefix %s", valstr);
+			    		free(pair);
+			    		return -1;
+			    	}
+			    	*p = 0;
+			    	p++;
+			    	pair->strvalue[0] = 0;
+			    	pair->strvalue[1] = atoi(p);
+
+			    	if (inet_pton(AF_INET6, valstr, pair->strvalue+2) == 0) {
+			    		rc_log(LOG_ERR, "rc_avpair_parse: invalid IPv6 prefix %s", valstr);
+			    		free(pair);
+			    		return -1;
+			    	}
+				pair->lvalue = 2+16;
+				break;
+
 			    case PW_TYPE_DATE:
 				timeval = time (0);
 				tm = localtime (&timeval);
@@ -645,7 +715,7 @@ int rc_avpair_parse (const rc_handle *rh, char *buffer, VALUE_PAIR **first_pair)
  *
  */
 
-int rc_avpair_tostr (const rc_handle *rh, VALUE_PAIR *pair, char *name, int ln, char *value, int lv)
+int rc_avpair_tostr (rc_handle const *rh, VALUE_PAIR *pair, char *name, int ln, char *value, int lv)
 {
 	DICT_VALUE     *dval;
 	char            buffer[32];
@@ -674,7 +744,7 @@ int rc_avpair_tostr (const rc_handle *rh, VALUE_PAIR *pair, char *name, int ln,
 		{
 			if (!(isprint (*ptr)))
 			{
-				sprintf (buffer, "\\%03o", *ptr);
+				snprintf (buffer, sizeof(buffer), "\\%03o", *ptr);
 				strncat(value, buffer, (size_t) lv);
 				lv -= 4;
 				if (lv < 0) break;
@@ -683,7 +753,7 @@ int rc_avpair_tostr (const rc_handle *rh, VALUE_PAIR *pair, char *name, int ln,
 			{
 				strncat(value, (char *)ptr, 1);
 				lv--;
-				if (lv < 0) break;
+				if (lv <= 0) break;
 			}
 			ptr++;
 		}
@@ -697,7 +767,7 @@ int rc_avpair_tostr (const rc_handle *rh, VALUE_PAIR *pair, char *name, int ln,
 		}
 		else
 		{
-			sprintf (buffer, "%ld", (long int)pair->lvalue);
+			snprintf(buffer, sizeof(buffer), "%ld", (long int)pair->lvalue);
 			strncpy(value, buffer, (size_t) lv);
 		}
 		break;
@@ -707,6 +777,26 @@ int rc_avpair_tostr (const rc_handle *rh, VALUE_PAIR *pair, char *name, int ln,
 		strncpy (value, inet_ntoa (inad), (size_t) lv-1);
 		break;
 
+	    case PW_TYPE_IPV6ADDR:
+	    	if (inet_ntop(AF_INET6, pair->strvalue, value, lv-1) == NULL)
+	    		return -1;
+		break;
+
+	    case PW_TYPE_IPV6PREFIX: {
+	    	uint8_t ip[16];
+	    	uint8_t txt[48];
+	    	if (pair->lvalue < 2)
+	    		return -1;
+
+	    	memset(ip, 0, sizeof(ip));
+	    	memcpy(ip, pair->strvalue+2, pair->lvalue-2);
+
+	    	if (inet_ntop(AF_INET6, ip, txt, sizeof(txt)) == NULL)
+	    		return -1;
+		snprintf(value, lv-1, "%s/%u", txt, (unsigned)pair->strvalue[1]);
+
+		break;
+	    }
 	    case PW_TYPE_DATE:
 		strftime (buffer, sizeof (buffer), "%m/%d/%y %H:%M:%S",
 			  gmtime ((time_t *) & pair->lvalue));
@@ -726,18 +816,16 @@ int rc_avpair_tostr (const rc_handle *rh, VALUE_PAIR *pair, char *name, int ln,
  * Function: rc_avpair_log
  *
  * Purpose: format sequence of attribute value pairs into printable
- * string. The string is dynamically allocated and is automatically
- * freed when rc_handle is destroyed or rc_avpair_log() is called
- * again.
+ * string. The caller should provide a storage buffer and the buffer length.
+ * Returns pointer to provided storage buffer.
  *
  */
 char *
-rc_avpair_log(rc_handle *rh, VALUE_PAIR *pair)
+rc_avpair_log(rc_handle const *rh, VALUE_PAIR *pair, char *buf, size_t buf_len)
 {
 	size_t len, nlen;
 	VALUE_PAIR *vp;
 	char name[33], value[256];
-	char *cp;
 
 	len = 0;
 	for (vp = pair; vp != NULL; vp = vp->next) {
@@ -745,16 +833,12 @@ rc_avpair_log(rc_handle *rh, VALUE_PAIR *pair)
 		    sizeof(value)) == -1)
 		        return NULL;
 		nlen = len + 32 + 3 + strlen(value) + 2 + 2;
-		cp = realloc(rh->ppbuf, nlen);
-		if (cp == NULL) {
-			rc_log(LOG_ERR, "rc_avpair_log: can't allocate memory");
-			return NULL;
-		}
-		sprintf(cp + len, "%-32s = '%s'\n", name, value);
-		rh->ppbuf = cp;
+		if(nlen<buf_len-1) {
+			sprintf(buf + len, "%-32s = '%s'\n", name, value);
+		} else return buf;
 		len = nlen - 1;
 	}
-	return (len > 0) ? rh->ppbuf : NULL;
+	return buf;
 }
 
 /*
@@ -765,7 +849,7 @@ rc_avpair_log(rc_handle *rh, VALUE_PAIR *pair)
  *
  */
 
-VALUE_PAIR *rc_avpair_readin(const rc_handle *rh, FILE *input)
+VALUE_PAIR *rc_avpair_readin(rc_handle const *rh, FILE *input)
 {
 	VALUE_PAIR *vp = NULL;
 	char buffer[1024], *q;
